# TabNN LightningModule
The `TabNNModule` Class is a Pytorch Lightning Module and the full recipe that defines how the TabNN model interacts. The base principles are described [here](https://lightning.ai/docs/pytorch/stable/common/lightning_module.html)

In contrast to the`SimpleNNRegInterface` implementation it has the following properties:

- It is independent of the `**config` for the `NNAlgInterface`. New configs can be defined using the dictionary in `TabNNModule.interface_config`.
- It makes the `Learner` Class obsolete, since it - in combination with the pytorch Lightning Trainer - handles the training, validation and prediction process.
- Probably (not implemented yet) it also makes the `LearnerProgress` and maybe also the `Schedule` obsolete, since the Trainer already keeps track of most of the information needed.
- It is responsible to create and initialize the optimizer. (no longer the `NNAlgInterface`). This optimizer is contentwise the same as the Adam optimizer from this repo, but using `pytorch.optim.Optimizer` base class. This is needed for the Lightning optimizer scheduler to work.
- The `NNAlgInterface` is used but in a different way than in `SimpleNNRegInterface`. Here we use a light version of the interface called `NNCreatorInterface` (the name interface makes probably no longer sense here ..).
This light version is still the core of the `TabNNModule` but it is only needed at initialization time to create the model and other components. Everything that happens during the training is handled by the Module or the Callbacks.

## Hooks and Callbacks
The original Callbacks from `callbacks.py` were reimplemented and simplified, using Pytorch Lightning Callbacks.
The Validation callback however is mostly implemented directly in the `TabNNModule` and 
only checkpointing and `restore_best` are outsourced in the `ModelCheckpointCallback`Class.

The callback hooks from pytorch lightning are similar to the original ones. They can be looked up [here](https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#hooks)