from tab_bench.data.paths import Paths
from tab_bench.data.tasks import TaskPackage, TaskDescription
from tab_bench.run.results import ResultManager
from tab_bench.scheduling.resources import NodeResources
from tab_models.sklearn.default_params import DefaultParams
from tab_models.training.logging import StdoutLogger
import custom_paths
import time
from tab_models.model.models import *
from tab_bench.alg_wrappers.interface_wrappers import NNInterfaceWrapper, MLPRTDLInterfaceWrapper, \
    ResNetRTDLInterfaceWrapper, TabrInterfaceWrapper
from tab_models.training.metrics import Metrics


def run_example(paths: Paths):
    start_time = time.time()
    use_gpu = torch.cuda.is_available()

    wrapper = NNInterfaceWrapper(**DefaultParams.MLP_TD_S_CLASS)

    # task_info = TaskDescription('uci-multi-class', 'chess_krvk').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'madelon').load_info(paths)
    # task_info = TaskDescription('uci-multi-class', 'firm_teacher_clave').load_info(paths)
    # task_info = TaskDescription('uci-multi-class', 'avila').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'first_order_theorem_proving').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'mushroom').load_info(paths)
    task_info = TaskDescription('uci-bin-class', 'eeg_eye_state').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'wall_follow_robot_24').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'sml2010_room').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'wine_quality_white').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'naval_propulsion_turb').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'indoor_loc_long').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'facebook_comment_volume').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'ozone_level_8hr').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'adult').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'waveform').load_info(paths)
    # task_info = TaskDescription('uci-multi-class', 'gesture_phase_segmentation_raw').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'firm_teacher_clave').load_info(paths)
    # task_info = TaskDescription('uci-multi-class', 'skill_craft').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'bank_marketing').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'parkinson_motor').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'wall_follow_robot_24').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'appliances_energy').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'bejing_pm25').load_info(paths)
    # task_info = TaskDescription('openml-class', 'amazon-commerce-reviews').load_info(paths)
    # task_info = TaskDescription('openml-class', 'dionis').load_info(paths)
    # task_info = TaskDescription('openml-reg', 'Allstate_Claims_Severity').load_info(paths)
    # task_info = TaskDescription('openml-reg', 'Airlines_DepDelay_10M').load_info(paths)
    # task_info = TaskDescription('openml-reg', 'nyc-taxi-green-dec-2016').load_info(paths)
    # task_info = TaskDescription('openml-reg', 'Allstate_Claims_Severity').load_info(paths)
    # task_info = TaskDescription('openml-reg', 'Buzzinsocialmedia_Twitter').load_info(paths)
    # task_info = TaskDescription('openml-class', 'amazon-commerce-reviews').load_info(paths)
    # task_info = TaskDescription('openml-reg', 'fifa').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'metro_interstate_traffic_volume_long').load_info(paths)
    # task_info = TaskDescription('uci-reg', 'insurance_benchmark').load_info(paths)
    print('n_samples:', task_info.n_samples)
    print('n_cont:', task_info.tensor_infos['x_cont'].get_n_features())
    print('x_cat cat sizes:', task_info.tensor_infos['x_cat'].get_cat_sizes())
    print('n_classes:', task_info.tensor_infos['y'].get_cat_sizes())
    if task_info.tensor_infos['y'].get_cat_sizes() > 0:
        class_frequencies = torch.bincount(task_info.load_task(paths).ds.tensors['y'].squeeze(-1))
        print(f'class frequencies: {class_frequencies.numpy()}')

    is_nn = (isinstance(wrapper, NNInterfaceWrapper) or isinstance(wrapper, MLPRTDLInterfaceWrapper)
             or isinstance(wrapper, ResNetRTDLInterfaceWrapper)
             or isinstance(wrapper, TabrInterfaceWrapper))
    use_gpu = use_gpu and is_nn

    print(f'Running on task {task_info.task_desc}')
    if is_nn:
        split_infos = task_info.get_random_splits(10)[0:1]
        task_package = TaskPackage(task_info, split_infos=split_infos, n_cv=1, n_refit=0, paths=paths, rerun=False,
                                   alg_name='test', save_y_pred=False)
    else:
        split_infos = task_info.get_random_splits(10)[1:2]
        task_package = TaskPackage(task_info, split_infos=split_infos, n_cv=1, n_refit=0, paths=paths, rerun=True,
                                   alg_name='test', save_y_pred=False)
    logger = StdoutLogger(verbosity_level=2)
    metric_name = Metrics.default_metric_name(task_info.task_type)
    required_resources = wrapper.get_required_resources(task_package)
    print(f'Predicted time usage in s: {required_resources.time_s:g}')
    print(f'Predicted CPU RAM usage in GB: {required_resources.cpu_ram_gb:g}')
    print(f'Requested n_threads: {required_resources.n_threads:g}')

    # metric_name = '1-auroc'
    gpu_usages = np.array([1.0]) if use_gpu and is_nn else np.array([], dtype=np.float32)
    gpu_rams_gb = np.array([5.0]) if use_gpu and is_nn else np.array([], dtype=np.float32)
    tmp_folders = [paths.results_alg_task_split(task_package.task_info.task_desc,
                                                alg_name=task_package.alg_name, n_cv=task_package.n_cv,
                                                split_type=split_info.split_type,
                                                split_id=split_info.id) / 'tmp' for split_info in
                   task_package.split_infos]
    result_managers = wrapper.run(task_package, logger,
                                  assigned_resources=NodeResources(node_id=0, n_threads=16.0, cpu_ram_gb=2.0,
                                                                   gpu_usages=gpu_usages,
                                                                   gpu_rams_gb=gpu_rams_gb,
                                                                   physical_core_usages=np.array([0.0])),
                                  tmp_folders=tmp_folders)

    for rm in result_managers:
        print(rm.metrics_dict)
        print(rm.other_dict)

    result_pairs = [('val', [rm.metrics_dict['cv']['val']['1']['0'][metric_name] for rm in result_managers])]
    for is_cv in [True, False] if task_package.n_refit > 0 else [True]:
        cv_str = 'cv' if is_cv else 'refit'
        max_n_models = task_package.n_cv if is_cv else task_package.n_refit
        for n_models in {1, max_n_models}:  # use a set in case max_n_models == 1
            try:
                name = 'test-' + cv_str + '-' + str(n_models)
                results = [rm.metrics_dict[cv_str]['test'][str(n_models)][str(start_idx)][metric_name]
                           for rm in result_managers for start_idx in range(1 if n_models > 1 else max_n_models)]
                result_pairs.append((name, results))
            except KeyError as e:
                print(e)
                pass  # might happen if wrapper is not a randomized alg and therefore does not do ensembling
    for name, results in result_pairs:
        print(f'Mean {name} error: {np.mean(results):g} +- {np.std(results) / np.sqrt(len(results)):g}')

    # for rm in rms:
    #     print('val:', rm.val_dict)
    #     print('test:', rm.test_dict)

    print(f'Time: {time.time() - start_time:g} s')


if __name__ == '__main__':
    run_example(Paths(base_folder=custom_paths.get_base_folder()))
