import pathlib
from typing import List, Optional, Union

import numpy as np

from tab_models import utils
from tab_models.alg_interfaces.alg_interfaces import AlgInterface, SingleSubSplitAlgInterface
from tab_models.alg_interfaces.classification_for_quantile_regression import ClassificationForRegressionAlgInterface
from tab_models.alg_interfaces.sub_split_interfaces import SingleSplitWrapperAlgInterface
from tab_models.alg_interfaces.xgboost_interfaces import XGBSubSplitInterface
from tab_models.sklearn.default_params import DefaultParams
from tab_models.sklearn.sklearn_base import AlgInterfaceRegressor


class XGBClassQuantileRegressor(AlgInterfaceRegressor):
    def __init__(self, device: Optional[str] = None, random_state: Optional[Union[int, np.random.RandomState]] = None,
                 n_cv: int = 1, n_refit: int = 0, val_fraction: float = 0.2, n_threads: Optional[int] = None,
                 tmp_folder: Optional[Union[str, pathlib.Path]] = None, verbosity: int = 0,
                 train_metric_name: Optional[str] = None, val_metric_name: Optional[str] = None,
                 n_estimators: Optional[int] = None,
                 max_depth: Optional[int] = None,
                 lr: Optional[float] = None,
                 subsample: Optional[float] = None,
                 colsample_bytree: Optional[float] = None,
                 colsample_bylevel: Optional[float] = None,
                 colsample_bynode: Optional[float] = None,
                 min_child_weight: Optional[float] = None,
                 alpha: Optional[float] = None,
                 reg_lambda: Optional[float] = None,
                 gamma: Optional[float] = None,
                 tree_method: Optional[str] = None,
                 max_delta_step: Optional[float] = None,
                 max_cat_to_onehot: Optional[int] = None,
                 num_parallel_tree: Optional[int] = None,
                 max_bin: Optional[int] = None,
                 multi_strategy: Optional[str] = None,
                 n_bins: Optional[int] = None,
                 quantile_alpha: Optional[float] = None,
                 binning_strategy: Optional[float] = None,
                 ):
        self.device = device
        self.random_state = random_state
        self.n_cv = n_cv
        self.n_refit = n_refit
        self.val_fraction = val_fraction
        self.n_threads = n_threads
        self.tmp_folder = tmp_folder
        self.verbosity = verbosity
        self.train_metric_name = train_metric_name
        self.val_metric_name = val_metric_name
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.lr = lr
        self.subsample = subsample
        self.colsample_bytree = colsample_bytree
        self.colsample_bylevel = colsample_bylevel
        self.colsample_bynode = colsample_bynode
        self.min_child_weight = min_child_weight
        self.alpha = alpha
        self.reg_lambda = reg_lambda
        self.gamma = gamma
        self.tree_method = tree_method
        self.max_delta_step = max_delta_step
        self.max_cat_to_onehot = max_cat_to_onehot
        self.num_parallel_tree = num_parallel_tree
        self.max_bin = max_bin
        self.multi_strategy = multi_strategy
        self.n_bins = n_bins
        self.quantile_alpha = quantile_alpha
        self.binning_strategy = binning_strategy

    def _get_default_params(self):
        return utils.join_dicts(DefaultParams.XGB_TD_CLASS,
                                dict(val_metric_name='cross_entropy', quantile_alpha=0.95, n_bins=20,
                                     binning_strategy='quantile', multi_strategy='multi_output_tree'))

    def _create_alg_interface(self, n_cv: int) -> AlgInterface:
        class_interface = SingleSplitWrapperAlgInterface([XGBSubSplitInterface(**self.get_config()) for i in range(n_cv)])
        return ClassificationForRegressionAlgInterface(classification_alg_interface=class_interface,
                                                       **self.get_config())

    def _allowed_device_names(self) -> List[str]:
        return ['cpu', 'cuda']
